#include "AST.h"
#include "ByteCode.h"

int ls::AST_Ident::generateCode(GenerationContext & t)
{
	t.push(code_t::push, BoxedVar(name));
	t.push(code_t::load);
	return 2;
}

int ls::AST_SentenceList::generateCode(GenerationContext & t)
{
	int ret = 0;
	for (auto & it : sentences) {
		ret += it->generateCode(t);
	}
	return ret;
}

int ls::AST_CommonSentence::generateCode(GenerationContext & t)
{
	int ret = expression->generateCode(t);
	t.push(code_t::pop);
	return ret + 1;
}

int ls::AST_VarSentence::generateCode(GenerationContext & t)
{
	int ret = RHS->generateCode(t);
	t.push(code_t::push, BoxedVar(LHS->name));
	t.push(code_t::new_var);
	return ret + 2;
}

int ls::AST_AssignmentSentence::generateCode(GenerationContext & t)
{
	int ret = RHS->generateCode(t);
	ret += LHS->generateCode(t);
	t.push(code_t::store);
	return ret + 1;
}

int ls::AST_Binary::generateCode(GenerationContext & t)
{
	int ret = RHS->generateCode(t);
	ret += LHS->generateCode(t);
	t.push(static_cast<code_t>(op));
	return ret + 1;
}

int ls::AST_Unary::generateCode(GenerationContext & t)
{
	int ret = LHS->generateCode(t);
	t.push(static_cast<code_t>(op));
	return ret + 1;
}

int ls::AST_While::generateCode(GenerationContext & t)
{
	t.loopStart();
	int ret = condition->generateCode(t);
	t.push(code_t::push, BoxedVar());
	auto e = t.currentPos();
	t.push(code_t::jf);
	ret += 2;
	auto r = ret;
	t.push(code_t::enterScope);
	ret++;
	ret += sentences->generateCode(t);
	t.push(code_t::leaveScope);
	ret++;
	ret++;
	t.push(code_t::push, BoxedVar(-ret));
	t.push(code_t::jmp);
	ret++;
	t.loopEnd();
	t.set(e, BoxedVar(ret + 1 - r));
	return ret;
}

int ls::AST_If::generateCode(GenerationContext & t)
{
	int ret = condition->generateCode(t);
	ret++;
	t.push(code_t::push, BoxedVar());
	auto r = ret;
	auto e = t.currentPos();
	t.push(code_t::jf);
	t.push(code_t::enterScope);
	ret += 3;
	ret += ifSentence->generateCode(t);
	t.push(code_t::leaveScope);
	ret++;
	t.set(e, BoxedVar(ret - r + (elseSentence != nullptr ? 1 : 0) /* ɾ */));
	if (elseSentence != nullptr) {
		t.push(code_t::jmp, BoxedVar());
		auto cur = ret;
		auto code_operand_address = t.currentPos();
		t.push(code_t::enterScope);
		ret++;
		ret += elseSentence->generateCode(t);
		t.push(code_t::leaveScope);
		ret++;
		t.set(code_operand_address, BoxedVar(ret - cur));
	}
	return ret;
}

int ls::AST_Dot::generateCode(GenerationContext & t)
{
	int ret = 2;
	t.push(code_t::push, BoxedVar(RHS->name));
	ret += LHS->generateCode(t);
	t.push(code_t::ref);
	return ret;
}

int ls::AST_Index::generateCode(GenerationContext & t)
{
	int ret = 1;
	ret += RHS->generateCode(t);
	ret += LHS->generateCode(t);
	t.push(code_t::ref);
	return ret;
}

int ls::AST_ForIn::generateCode(GenerationContext & t)
{
	int ret = 1;
	t.push(code_t::enterScope);
	ret += iterable->generateCode(t);
	t.push(code_t::getiter);
	t.push(code_t::push, BoxedVar());
	t.push(code_t::push, BoxedVar(id->name));
	t.push(code_t::new_var);
	ret += 4;
	t.loopStart();
	{
		t.push(code_t::isEnd);
		t.push(code_t::push, BoxedVar(3));
		t.push(code_t::jf);
		t.registerBreak();
		t.push(code_t::getvalue);
		t.push(code_t::push, BoxedVar(id->name));
		t.push(code_t::store);
		ret += 8;
		ret += sentences->generateCode(t);
		t.push(code_t::next);
		ret++;
	}
	ret += t.registerContinue();
	t.push(code_t::pop);
	t.push(code_t::leaveScope);
	ret += 2;
	t.loopEnd();
	return ret;
}

int ls::AST_CallExpr::generateCode(GenerationContext & t)
{
	int ret = 0;
	int argnum = 0;
	for (auto it = args.rbegin(); it != args.rend(); it++) {
		ret += (*it)->generateCode(t);
		argnum++;
	}
	t.push(code_t::push, BoxedVar(argnum));
	ret++;
	ret += callee->generateCode(t);
	t.push(code_t::call);
	return ret + 1;
}

int ls::AST_Break::generateCode(GenerationContext & t)
{
	return t.registerBreak();
}

int ls::AST_Continue::generateCode(GenerationContext & t)
{
	return t.registerContinue();
}

int ls::AST_Function::generateCode(GenerationContext & t)
{	//ϴ    ̷ «.
	static int64_t num=0;
	auto c = std::unique_ptr<FunctionCode, FunctionCodeDeleter>(new FunctionCode);
	auto presentFrame = t.getFrame();
	t.setFrame(c.get());
	{
		t.push(code_t::pop);
		t.push(code_t::enterScope);
		for (auto & it : args){
			t.push(code_t::push, BoxedVar(it->name));
			t.push(code_t::new_var);
		}
		sentences->generateCode(t);
	}
	t.setFrame(presentFrame);
	t.push(code_t::push, BoxedVar( std::to_string(num++), move(c) )
	);
	return 1;
}

int ls::AST_Return::generateCode(GenerationContext & t)
{
	int ret = return_value->generateCode(t);
	t.push(code_t::ret);
	return ret + 1;
}
