#include<string>
#include<memory>
#include<ostream>
#include<cstdint>
#include<list>

#include"Operators.h"
#include"CompoundType.h"
#include"CodeGeneration.h"
#include"LuaScriptGeneration.h"

#ifndef LS_AST_H_
#define LS_AST_H_
namespace ls {
	using std::move;

	class AST_Sentence
	{
	public:
		virtual int generateCode(GenerationContext &) = 0;
		virtual void printDebugInfo(std::ostream &) = 0;
		virtual void luaCodeGen(LuaScriptGenerateContext &) = 0;
		virtual ~AST_Sentence() {}
	};
	class AST_SentenceList : public AST_Sentence {
	public:
		virtual int generateCode(GenerationContext &);
		void push_back(std::unique_ptr<AST_Sentence> s) {
			sentences.emplace_back(move(s));
		}
		virtual void printDebugInfo(std::ostream & os) override {
			for (auto & it : sentences) {
				it->printDebugInfo(os);
			}
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			for (auto & it : sentences)
			{
				lsgc.sentence_tap();
				it->luaCodeGen(lsgc);
			}
		}
	private:
		std::list< std::unique_ptr<AST_Sentence> > sentences;
	};
	class AST_Expr {
	public:
		virtual int generateCode(GenerationContext &) = 0;
		virtual void luaCodeGen(LuaScriptGenerateContext &) = 0;
		//Twin String . ߿ ٲ۴.
		virtual void printDebugInfo(std::ostream &) = 0;
		virtual ~AST_Expr() {}
	};
	//Lvalue ǥ    ̷.
	class AST_Variable : public AST_Expr {};

	class AST_Ident : public AST_Variable {
	public:
		explicit AST_Ident(const std::string & s) :name(s) {}
		virtual int generateCode(GenerationContext &) override;

		virtual void printDebugInfo(std::ostream & os) {
			os << name;
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			lsgc.out() << name;
		}

		std::string name;
	};
	inline std::ostream & operator<<(std::ostream & os, std::unique_ptr<AST_Sentence> & rhs) {
		rhs->printDebugInfo(os);
		return os;
	}
	inline std::ostream & operator<<(std::ostream & os, std::unique_ptr<AST_SentenceList> & rhs) {
		rhs->printDebugInfo(os);
		return os;
	}
	inline std::ostream & operator<<(std::ostream & os, std::unique_ptr<AST_Expr> & rhs) {
		rhs->printDebugInfo(os);
		return os;
	}
	inline std::ostream & operator<<(std::ostream & os, std::unique_ptr<AST_Variable> & rhs) {
		rhs->printDebugInfo(os);
		return os;
	}
	inline std::ostream & operator<<(std::ostream & os, std::unique_ptr<AST_Ident> & rhs) {
		rhs->printDebugInfo(os);
		return os;
	}
	
	class AST_CommonSentence : public AST_Sentence {
	public:
		explicit AST_CommonSentence(std::unique_ptr<AST_Expr> expr) :expression(move(expr)) {}
		virtual int generateCode(GenerationContext &);
		virtual void printDebugInfo(std::ostream & os) override {
			os << expression << "\n";
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			expression->luaCodeGen(lsgc);
			lsgc.out() << "\n";
		}
	private:
		std::unique_ptr<AST_Expr> expression;
	};
	class AST_VarSentence : public AST_Sentence {
	public:
		AST_VarSentence(std::unique_ptr<AST_Ident> lhs, std::unique_ptr<AST_Expr> rhs)
			:LHS(move(lhs)), RHS(move(rhs)) {}
		virtual int generateCode(GenerationContext & t);
		virtual void printDebugInfo(std::ostream & os) override {
			os  <<"var "<< LHS <<" = " << RHS << "\n";
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			if (lsgc.depth > 0){
				lsgc.out() << "local ";
			}
			lsgc.out() << LHS->name << " = ";
			RHS->luaCodeGen(lsgc);
			lsgc.out() << "\n";
		}
	private:
		std::unique_ptr<AST_Ident> LHS;
		std::unique_ptr<AST_Expr> RHS;
	};
	class AST_AssignmentSentence : public AST_Sentence {
	public:
		AST_AssignmentSentence(std::unique_ptr<AST_Expr> lhs, std::unique_ptr<AST_Expr> rhs)
			:LHS(move(lhs)), RHS(move(rhs)) {}
		virtual int generateCode(GenerationContext & t);
		virtual void printDebugInfo(std::ostream & os) override {
			os << LHS << " = " << RHS << "\n";
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			LHS->luaCodeGen(lsgc);
			lsgc.out() << " = ";
			RHS->luaCodeGen(lsgc);
			lsgc.out() << "\n";
		}
	private:
		std::unique_ptr<AST_Expr> LHS;
		std::unique_ptr<AST_Expr> RHS;
	}; 
	class AST_If : public AST_Sentence {
	public:
		explicit AST_If(std::unique_ptr<AST_Expr> cond, std::unique_ptr<AST_SentenceList> s)
			:condition(move(cond)), ifSentence(move(s)),elseSentence(nullptr) {}
		virtual int generateCode(GenerationContext &);
		void setElse(std::unique_ptr<AST_SentenceList> e) { elseSentence = move(e); }
		virtual void printDebugInfo(std::ostream & os) override {
			os << "if " << condition << " then\n";
			os << ifSentence;
			if (elseSentence != nullptr)
				os << "else\n" << elseSentence;
			os << "end\n";
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			lsgc.out() << "if ";
			condition->luaCodeGen(lsgc);
			lsgc.out() << " then\n";
			lsgc.depth++;
			ifSentence->luaCodeGen(lsgc);
			if (elseSentence != nullptr) {
				lsgc.out() << "else\n";
				elseSentence->luaCodeGen(lsgc);
			}
			lsgc.out() << "end\n";
			lsgc.depth--;
		}
	private:
		std::unique_ptr<AST_Expr> condition;
		std::unique_ptr<AST_SentenceList> ifSentence;
		std::unique_ptr<AST_SentenceList> elseSentence;
	};
	class AST_While : public AST_Sentence {
	public:
		explicit AST_While(std::unique_ptr<AST_Expr> cond, std::unique_ptr<AST_SentenceList> s)
			:condition(move(cond)),sentences(move(s)) {}
		virtual int generateCode(GenerationContext &);
		virtual void printDebugInfo(std::ostream & os) override {
			os << "while " << condition << " do\n";
			sentences->printDebugInfo(os);
			os << "end\n";
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			lsgc.out() << "while ";
			condition->luaCodeGen(lsgc);
			lsgc.out() << " do\n";
			lsgc.depth++;
			sentences->luaCodeGen(lsgc);
			lsgc.out() << "end\n";
			lsgc.depth--;
		}
	private:
		std::unique_ptr<AST_Expr> condition;
		std::unique_ptr<AST_SentenceList> sentences;
	};
	class AST_ForIn : public AST_Sentence {
	public:
		explicit AST_ForIn(std::unique_ptr<AST_Ident> i,std::unique_ptr<AST_Expr> cond,
			std::unique_ptr<AST_SentenceList> s) 
			:id(move(i)),iterable(move(cond)),sentences(move(s)) {}
		virtual int generateCode(GenerationContext &);
		virtual void printDebugInfo(std::ostream & os) override {
			os << "for " << id << " in " << iterable << " do\n";
			sentences->printDebugInfo(os);
			os << "end\n";
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			lsgc.out() << "for " << id->name << ", _ in pairs(";
			iterable->luaCodeGen(lsgc);
			lsgc.out() << ") do\n";
			lsgc.depth++;
			sentences->luaCodeGen(lsgc);
			lsgc.out() << "end\n";
			lsgc.depth--;
		}
	private:
		std::unique_ptr<AST_Ident> id;
		std::unique_ptr<AST_Expr> iterable;
		std::unique_ptr<AST_SentenceList> sentences;
	};
	class AST_Break : public AST_Sentence {
	public:
		explicit AST_Break() {}
		virtual int generateCode(GenerationContext &);
		virtual void printDebugInfo(std::ostream & os) override {
			os << "break\n";
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			lsgc.out() << "break\n";
		}
	}; 
	class AST_Continue : public AST_Sentence {
	public:
		explicit AST_Continue() {}
		virtual int generateCode(GenerationContext &);
		virtual void printDebugInfo(std::ostream & os) override {
			os << "continue\n";
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			lsgc.out() << "continue\n";
		}
	};
	class AST_Return : public AST_Sentence {
	public:
		explicit AST_Return(std::unique_ptr<AST_Expr> r)
		:return_value(move(r)){}
		virtual int generateCode(GenerationContext &);
		virtual void printDebugInfo(std::ostream & os) override {
			os << "return " << return_value << "\n";
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			lsgc.out() << "return ";
			return_value->luaCodeGen(lsgc);
			lsgc.out() << "\n";
		}

		std::unique_ptr<AST_Expr> return_value;
	};
	class AST_Binary : public AST_Expr {
	public:
		explicit AST_Binary(Opers oper, std::unique_ptr<AST_Expr> l, std::unique_ptr<AST_Expr> r)
			:op(oper), LHS(move(l)), RHS(std::move(r)) {}
		virtual int generateCode(GenerationContext &);
		virtual void printDebugInfo(std::ostream & os) {
			os << LHS << " " << to_string(op) << " " << RHS;
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			LHS->luaCodeGen(lsgc);
			lsgc.out() << to_string(op);
			RHS->luaCodeGen(lsgc);
		}
	protected:
		Opers op;
		std::unique_ptr<AST_Expr> LHS, RHS;
	};
	class AST_Unary : public AST_Expr {
	public:
		explicit AST_Unary(Opers oper, std::unique_ptr<AST_Expr> l) :op(oper)
			, LHS(move(l)) {}
		virtual int generateCode(GenerationContext &);
		virtual void printDebugInfo(std::ostream & os) {
			os << to_string(op) << " " << LHS;
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			if(op != Opers::kPlus)
				lsgc.out() << to_string(op);
			LHS->luaCodeGen(lsgc);
		}
	protected:
		Opers op;
		std::unique_ptr<AST_Expr> LHS;
	};
	class AST_Index : public AST_Variable {
	public:
		explicit AST_Index(std::unique_ptr<AST_Expr> l, std::unique_ptr<AST_Expr> r)
			:LHS(move(l)), RHS(std::move(r)) {}
		virtual int generateCode(GenerationContext &);
		virtual void printDebugInfo(std::ostream & os) {
			os << LHS << "[" << RHS << "]";
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			LHS->luaCodeGen(lsgc);
			lsgc.out() << "[";
			RHS->luaCodeGen(lsgc);
			lsgc.out() << "]";
		}
	protected:
		std::unique_ptr<AST_Expr> LHS, RHS;
	};
	class AST_Dot : public AST_Variable {
	public:
		explicit AST_Dot(std::unique_ptr<AST_Expr> l, std::unique_ptr<AST_Ident> r)
			:LHS(move(l)), RHS(std::move(r)) {}
		virtual int generateCode(GenerationContext &);
		virtual void printDebugInfo(std::ostream & os) {
			os << LHS << "." << RHS;
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			LHS->luaCodeGen(lsgc);
			lsgc.out() << "." << RHS->name;
		}
	protected:
		std::unique_ptr<AST_Expr> LHS;
		std::unique_ptr<AST_Ident> RHS;
	};
	class AST_CallExpr : public AST_Variable {
	public:
		explicit AST_CallExpr(std::unique_ptr<AST_Expr> l)
			:callee(move(l)){}
		void push_back(std::unique_ptr<AST_Expr> arg) { args.emplace_back(move(arg)); }
		virtual int generateCode(GenerationContext &);
		virtual void printDebugInfo(std::ostream & os) {
			os << callee << "(";
			if (args.size() != 0) {
				auto it = args.begin();
				os << *it++;
				while (it != args.end()) {
					os << "," << *it++;
				}
			}
			os << ")";
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			lsgc.out() << callee << "(";
			if (args.size() != 0) {
				auto it = args.begin();
				(*it++)->luaCodeGen(lsgc);
				while (it != args.end()) {
					lsgc.out() << ",";
					(*it++)->luaCodeGen(lsgc);
				}
			}
			lsgc.out() << ")";
		}
	protected:
		std::unique_ptr<AST_Expr> callee;
		std::list<std::unique_ptr<AST_Expr>> args;
	};
	class AST_Function : public AST_Expr {
	public:
		void push_back(std::unique_ptr<AST_Ident> arg) { args.emplace_back(move(arg)); }
		void setSentenceList(std::unique_ptr<AST_SentenceList> s) { sentences = move(s); }
		virtual int generateCode(GenerationContext &);
		virtual void printDebugInfo(std::ostream & os) {
			os << "fn " << "(";
			if (args.size() != 0) {
				auto it = args.begin();
				os << *it++;
				while (it != args.end()) {
					os << "," << *it++;
				}
			}
			os << ")\n";
			os << sentences << "end\n";
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			lsgc.out() << "function" << "(";
			
			if (args.size() != 0) {
				auto it = args.begin();
				lsgc.out() << (*it++)->name;
				while (it != args.end()) {
					lsgc.out() << "," << (*it++)->name;
				}
			}
			lsgc.out() << ")\n";
			lsgc.depth++;
			sentences->luaCodeGen(lsgc);
			lsgc.depth--;
			lsgc.out() << "end\n";
		}
	protected:
		std::list<std::unique_ptr<AST_Ident>> args;
		std::unique_ptr<AST_SentenceList> sentences;
	};
	
	template<typename Ty>
	class AST_Value : public AST_Expr {
	public:
		explicit AST_Value(const Ty & v) :value(v) {}
		virtual int generateCode(GenerationContext & gctx) override {
			gctx.push(code_t::push, BoxedVar(value));
			return 1;
		}
		virtual void printDebugInfo(std::ostream & os) {
			os << value;
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			lsgc.out() << value;
		}
	protected:
		Ty value;
	};

	class AST_Int : public AST_Value<int64_t> {
	public:
		using AST_Value<int64_t>::AST_Value;
	};
	class AST_Float : public AST_Value<double> {
	public:
		using AST_Value<double>::AST_Value;
	}; 
	class AST_None : public AST_Expr {
	public:
		virtual int generateCode(GenerationContext & gctx) override {
			gctx.push(code_t::push, BoxedVar());
			return 1;
		}
		virtual void printDebugInfo(std::ostream & os) {
			os << "None";
		}
		virtual void luaCodeGen(LuaScriptGenerateContext & lsgc) override {
			lsgc.out() << "nil";
		}
	};
}
#endif