#include "Parser.h"

using namespace ls;

using std::make_unique;

unique_ptr<AST_SentenceList> ls::Parser::parseSentenceList()
{
	auto ret = make_unique<AST_SentenceList>();
	for (;;) {
		auto sentence = parseSentence();
		if (sentence == nullptr)
			break;
		ret->push_back(move(sentence));
	}
	return ret;
}

unique_ptr<AST_Sentence> ls::Parser::parseSentence()
{
	unique_ptr<AST_Sentence> ret;
	if (it->str == "continue") {
		it++;
		ret = make_unique<AST_Continue>();
	}
	else if (it->str == "break") {
		it++;
		ret = make_unique<AST_Break>();
	}
	else if (it->str == "var") {
		ret = parseVarSentence();
	}
	else if (it->str == "return") {
		ret = parseReturn();
	}
	else if (it->str == "while") {
		ret = parseWhileSentence();
	}
	else if (it->str == "if") {
		ret = parseIfSentence();
	}
	else if (it->str == "for") {
		ret = parseForInSentence();
	}
	else {
		ret = parseCommonSentence();
	}
	return ret;
}

unique_ptr<AST_Sentence> ls::Parser::parseCommonSentence()
{
	unique_ptr<AST_Sentence> ret;
	auto lvalue = expr();
	if (lvalue == nullptr) return nullptr;
	if (it->str == "=") {
		it++;
		auto r = expr();
		return make_unique<AST_AssignmentSentence>(move(lvalue), move(r));
	}
	return make_unique<AST_CommonSentence>(move(lvalue));
}

unique_ptr<AST_Sentence> ls::Parser::parseVarSentence()
{
	if (it->str == "var") {
		it++;
		auto lv = parseIdentifier();
		if (lv == nullptr)
			ThrowExpected("Lvalue Expected");
		if (it->str != "=")
			ThrowExpected("expected symbol \'=\'");
		it++;
		auto r = expr();
		if (r == nullptr)
			ThrowExpected("Rvalue Expected");
		return (make_unique<AST_VarSentence>(move(lv), move(r)));
	}
	return nullptr;
}

unique_ptr<AST_Sentence> ls::Parser::parseWhileSentence()
{
	if (it->str == "while") {
		it++;
		auto lv = expr();
		if (lv == nullptr)
			ThrowExpected("expr Expected");
		if (it->str != "do")
			ThrowExpected("expected keyword \'do\'");
		it++;
		auto r = parseSentenceList();
		if (r == nullptr)
			ThrowExpected("Rvalue Expected");
		if (it->str != "end")
			ThrowExpected("end Expected");
		it++;
		return (make_unique<AST_While>(move(lv), move(r)));
	}
	return nullptr;
}

unique_ptr<AST_Sentence> ls::Parser::parseIfSentence()
{
	if (it->str == "if") {
		it++;
		auto lv = expr();
		if (lv == nullptr)
			ThrowExpected("expr Expected");
		if (it->str != "then")
			ThrowExpected("expected keyword \'then\'");
		it++;
		auto r = parseSentenceList();
		if (r == nullptr)
			ThrowExpected("Rvalue Expected");
		auto ret = (make_unique<AST_If>(move(lv), move(r)));
		if (it->str == "else") {
			it++;
			auto elsec = parseSentenceList();
			if (elsec == nullptr)
				ThrowExpected("expected Sentence next to else");
			ret->setElse(move(elsec));
		}
		if (it->str != "end")
			ThrowExpected("end Expected");
		return ret;
	}
	return nullptr;
}

unique_ptr<AST_Sentence> ls::Parser::parseForInSentence()
{
	if (it->str == "for") {
		it++;
		auto id = parseIdentifier();
		if (id == nullptr) {
			ThrowExpected("expected identifier");
		}
		if (it->str != "in")
			ThrowExpected("expected keyword \'in\'");
		it++;
		auto lv = expr();
		if (lv == nullptr)
			ThrowExpected("expr Expected");
		if (it->str != "do")
			ThrowExpected("expected keyword \'do\'");
		it++;
		auto sl = parseSentenceList();
		if (sl == nullptr)
			ThrowExpected("Rvalue Expected");
		if (it->str != "end")
			ThrowExpected("end Expected");
		return make_unique<AST_ForIn>(move(id), move(lv), move(sl));
	}
	return nullptr;
}

unique_ptr<AST_Sentence> ls::Parser::parseReturn()
{
	if (it->str == "return") {
		it++;
		auto retValue = expr();
		if (retValue != nullptr)
			return make_unique<AST_Return>(move(retValue));
		else
			return make_unique<AST_Return>(make_unique<AST_None>());
	}
	return nullptr;
}

unique_ptr<AST_Expr> Parser::parseLogicalBinary()
{
	unique_ptr<AST_Expr> LHS, RHS;
	LHS = parseLogicalUnary();
	if (LHS == nullptr)
		return nullptr;
	while (it->str == "and" || it->str == "or") {
		auto op = it->str;
		it++;
		RHS = parseLogicalUnary();
		if (RHS == nullptr)
			ThrowExpected("Expected Expression");
		if (op == "and")
			LHS = make_unique<AST_Binary>(kAnd,move(LHS), move(RHS));
		else if (op == "or")
			LHS = make_unique<AST_Binary>(kOr,move(LHS), move(RHS));
	}
	return LHS;
}

unique_ptr<AST_Expr> ls::Parser::parseLogicalUnary()
{
	if (it->str == "not") {
		++it; 
		auto ret = parseComparison();
		if (ret == nullptr)
			ThrowExpected("Expected Expression");
		return make_unique<AST_Unary>(kNot,move(ret));
	}
	return parseComparison();
}

unique_ptr<AST_Expr> ls::Parser::parseComparison()
{
	unique_ptr<AST_Expr> LHS, RHS;
	LHS = parseAddAndSub();
	if (LHS == nullptr)
		return nullptr;
	while (it->str == "=="||
		it->str == "!="||
		it->str == "<"||
		it->str == "<="||
		it->str == ">"||
		it->str == ">=") {
		auto op = it->str;
		++it;
		RHS = parseAddAndSub();
		if (RHS == nullptr)
			ThrowExpected("Expected Expression");
		if (op == "==")
			LHS = make_unique<AST_Binary>(kEqual, move(LHS), move(RHS));
		else if (op == "!=")
			LHS = make_unique<AST_Binary>(kNotEqual, move(LHS), move(RHS));
		else if (op == ">")
			LHS = make_unique<AST_Binary>(kGreater, move(LHS), move(RHS));
		else if (op == ">=")
			LHS = make_unique<AST_Binary>(kEqualGreater, move(LHS), move(RHS));
		else if (op == "<")
			LHS = make_unique<AST_Binary>(kLess, move(LHS), move(RHS));
		else if (op == "<=")
			LHS = make_unique<AST_Binary>(kLessEqual, move(LHS), move(RHS));
	}
	return LHS;
}

unique_ptr<AST_Expr> ls::Parser::parseAddAndSub()	//code 丵  .
{
	unique_ptr<AST_Expr> LHS, RHS;
	if (it->str == "+") {
		++it;
		LHS = parseMulAndDiv();
		if (LHS == nullptr)
			ThrowExpected("Expected Expression");
		LHS = make_unique<AST_Unary>(kPlus,move(LHS));
	}
	else if (it->str == "-") {
		++it;
		LHS = parseMulAndDiv();
		if (LHS == nullptr)
			ThrowExpected("Expected Expression");
		LHS = make_unique<AST_Unary>(kMinus,move(LHS));
	}
	else {
		LHS = parseMulAndDiv();
	}
	if (LHS == nullptr)
		return nullptr;
	while (it->str == "+" || it->str == "-") {
		auto op = it->str;
		++it;
		RHS = parseMulAndDiv();
		if (RHS == nullptr)
			ThrowExpected("Expected Expression");
		if (op == "+")
			LHS = make_unique<AST_Binary>(kAdd,move(LHS), move(RHS));
		else if (op == "-")
			LHS = make_unique<AST_Binary>(kSub,move(LHS), move(RHS));
	}
	return LHS;
}

unique_ptr<AST_Expr> ls::Parser::parseMulAndDiv()
{
	unique_ptr<AST_Expr> LHS, RHS;
	LHS = parseAttrAccess();
	if (LHS == nullptr)
		return nullptr;
	while (it->str == "*" || it->str == "/") {
		auto op = it->str;
		++it;
		RHS = parseAttrAccess();
		if (RHS == nullptr)
			ThrowExpected("Expected Expression");
		if (op == "*")
			LHS = make_unique<AST_Binary>(kMul,move(LHS), move(RHS));
		else if (op == "/")
			LHS = make_unique<AST_Binary>(kDiv,move(LHS), move(RHS));
	}
	return LHS;
}

unique_ptr<AST_Expr> ls::Parser::parseAttrAccess()
{
	unique_ptr<AST_Expr> LHS;
	LHS = parsePrimary();
	if (LHS == nullptr)
		return nullptr;
	for(;;){
		if (it->str == "[") {
			++it;
			auto RHS = expr();
			if (RHS == nullptr)
				ThrowExpected("Expected Expression");
			if (it->str != "]")
				ThrowExpected("Expected \')\'");
			it++;
			LHS = make_unique<AST_Index>(move(LHS), move(RHS));
		}
		else if (it->str == ".") {
			it++;
			auto RHS = parseIdentifier();
			LHS = make_unique<AST_Dot>(move(LHS), move(RHS));
		}
		else if (it->str == "(") {
			++it;
			auto L = make_unique<AST_CallExpr>(move(LHS));
			if (it->str != ")")
			{
				auto RHS = expr();
				if (RHS == nullptr) {
					ThrowExpected("Expected expression.");
				}
				L->push_back(move(RHS));
				while (it->str != ")") {
					if (it->str != ",") {
						ThrowExpected("Expected \',\'");
					}
					it++;
					RHS = expr();
					if (RHS == nullptr)
						ThrowExpected("Expected expression.");
					L->push_back(move(RHS));
				}
				it++;
			}
			LHS = move(L);
		}
		else break;
	}
	return LHS;
}

unique_ptr<AST_Expr> ls::Parser::parsePrimary()
{
	if (it->str == "(")
	{
		it++;
		auto ret = expr();
		if (ret == nullptr) {
			ThrowExpected("expected expression");
		}
		if (it->str != ")") {
			ThrowExpected("expected \')\'");
		}
		it++;
		return ret;
	}
	else if (it->type == ttIdentifier) {
		return parseIdentifier();
	}
	return parseLiteral();
}

unique_ptr<AST_Expr> ls::Parser::parseLiteral()
{
	if (it->type == ls::TokenType::ttInteger) {
		return parseInteger();
	}
	else if (it->type == ttFloat) {
		return parseFloating();
	}
	else if (it->str == "fn") {
		return parseFunction();
	}
	else if (it->str == "{") {
		return parseTableExpr();
	}
	return nullptr;
}

unique_ptr<AST_Expr> Parser::parseInteger()
{
	if (it->type != TokenType::ttInteger)
		return nullptr;
	auto value = std::stoll(it->str);
	it++;
	return make_unique<AST_Int>(value);
}

unique_ptr<AST_Expr> Parser::parseFloating()
{
	if (it->type != TokenType::ttFloat)
		return nullptr;
	auto value = std::stod(it->str);
	it++;
	return make_unique<AST_Float>(value);
}

unique_ptr<AST_Expr> ls::Parser::parseFunction()
{
	if (it->str != "fn")
		return nullptr;
	it++;
	if (it->str != "(")
		ThrowExpected("expected \'(\'");
	auto L = make_unique<AST_Function>();
	it++;
	if (it->str != ")")
	{
		auto Id = parseIdentifier();
		if (Id == nullptr) {
			ThrowExpected("Expected Identifier.");
		}
		L->push_back(move(Id));
		while (it->str != ")") {
			if (it->str != ",") {
				ThrowExpected("Expected \',\'");
			}
			it++;
			Id = parseIdentifier();
			if (Id == nullptr)
				ThrowExpected("Expected Identifier.");
			L->push_back(move(Id));
		}
		it++;
	}
	L->setSentenceList(parseSentenceList());
	if (it->str != "end")
		ThrowExpected("expected \'end\'");
	it++;
	return L;
}

unique_ptr<AST_Expr> ls::Parser::parseTableExpr()
{
	return nullptr;
}

unique_ptr<AST_Ident> ls::Parser::parseIdentifier()
{
	if (it->type == ttIdentifier) {
		auto ret = make_unique<AST_Ident>(it->str);
		it++;
		return ret;
	}
	return nullptr;
}
