﻿#include "VirtualMachine.h"

using namespace ls;

bool ls::VirtualMachine::runOneCode()
{
	if (programCounter>=programFrameEnd) {
		return false;
	}
	switch (programFrame->code[programCounter].code)
	{
	case code_t::nop:
		break;
	case code_t::add:
		add();
		break;
	case code_t::enterScope:
		enterScope();
		break;
	case code_t::leaveScope:
		leaveScope();
		break;
	case code_t::load:
		load();
		break;
	case code_t::and:
		and();
		break;
	case code_t::or:
		or();
		break;
	case code_t::not:
		not();
		break;
	case code_t::sub:
		sub();
		break;
	case code_t::mul:
		mul();
		break;
	case code_t::div:
		div();
		break;
	case code_t::plus:
		plus();
		break;
	case code_t::minus:
		minus();
		break;
	case code_t::eq:
		eq();
		break;
	case code_t::neq:
		neq();
		break;
	case code_t::gt:
		gt();
		break;
	case code_t::gte:
		gte();
		break;
	case code_t::less:
		less();
		break;
	case code_t::leq:
		leq();
		break;
	case code_t::push:
		push();
		break;
	case code_t::pop:
		pop();
		break;
	case code_t::store:
		store();
		break;
	case code_t::new_var:
		new_var();
		break;
	case code_t::jmp:
		jmp();
		break;
	case code_t::jt:
		jt();
		break;
	case code_t::jf:
		jf();
		break;
	case code_t::call:
		call();
		break;
	case code_t::ref:
		ref();
		break;
	case code_t::ret:
		inst_ret();
		break;
	default:
		throw std::runtime_error("Invailed op code.");
		break;
	}
	programCounter++;
	return true;
}

BoxedVar & ls::VirtualMachine::searchVar(const std::string & str)
{
	for (auto it = variableMap.rbegin(), E = variableMap.rend(); it != E; it++) {
		auto f = it->find(str);
		if(f != it->end()){
			return f->second;
		}
	}
	throw std::runtime_error("Unknown Variable");
}

void ls::VirtualMachine::enterScope()
{
	variableMap.emplace_back();
}

void ls::VirtualMachine::leaveScope()
{
	variableMap.pop_back();
}

void ls::VirtualMachine::add()
{
	auto LHS = popValue();
	auto RHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->binaryGo(Opers::kAdd, RHS.get()));
	pushValue(tmp);
}

void ls::VirtualMachine::load()
{
	auto str = popValue();
	if (str.get()->Type() != kString) {
		throw std::runtime_error("Load Failed!");
	}
	TempVariable t(&searchVar(static_cast<StringType *>(str.get())->value));
	pushValue(t);
}

void ls::VirtualMachine::sub()
{
	auto LHS = popValue();
	auto RHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->binaryGo(Opers::kSub, RHS.get()));
	pushValue(tmp);
}

void ls::VirtualMachine::mul()
{
	auto LHS = popValue();
	auto RHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->binaryGo(Opers::kMul, RHS.get()));
	pushValue(tmp);
}

void ls::VirtualMachine::div()
{
	auto LHS = popValue();
	auto RHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->binaryGo(Opers::kDiv, RHS.get()));
	pushValue(tmp);
}

void ls::VirtualMachine::plus()
{
	auto LHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->unaryGo(Opers::kPlus));
	pushValue(tmp);
}

void ls::VirtualMachine::minus()
{
	auto LHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->unaryGo(Opers::kMinus));
	pushValue(tmp);
}


void ls::VirtualMachine::eq()
{
	auto LHS = popValue();
	auto RHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->binaryGo(Opers::kEqual, RHS.get()));
	pushValue(tmp);
}

void ls::VirtualMachine::gt()
{
	auto LHS = popValue();
	auto RHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->binaryGo(Opers::kGreater, RHS.get()));
	pushValue(tmp);
}

void ls::VirtualMachine::gte()
{
	auto LHS = popValue();
	auto RHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->binaryGo(Opers::kEqualGreater, RHS.get()));
	pushValue(tmp);
}

void ls::VirtualMachine::less()
{
	auto LHS = popValue();
	auto RHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->binaryGo(Opers::kLess, RHS.get()));
	pushValue(tmp);
}

void ls::VirtualMachine::leq()
{
	auto LHS = popValue();
	auto RHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->binaryGo(Opers::kLessEqual, RHS.get()));
	pushValue(tmp);
}

void ls::VirtualMachine::neq()
{
	auto LHS = popValue();
	auto RHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->binaryGo(Opers::kNotEqual, RHS.get()));
	pushValue(tmp);
}

void ls::VirtualMachine::new_var()
{
	auto lst = popValue();
	auto rst = popValue();
	if (lst.get()->Type() != kString)
	{
		throw std::runtime_error("변수이름은 문자열이어야 합니다.");
	}

	auto name = static_cast<StringType *>(lst.get())->value;
	auto pos = variableMap.back().find(name);
	if (pos != variableMap.back().end()) {
		throw std::runtime_error("이미 선언된 변수.");
	}
	variableMap.back().emplace(name, *rst.val());
}

void ls::VirtualMachine::jmp()
{
	auto st = popValue();
	auto s = st.cast<IntegerType>(kInteger);
	programCounter += s->value;
	programCounter--;
}

void ls::VirtualMachine::jt()
{
	auto jmp_count = popValue();
	auto condition = popValue();
	auto jmp = jmp_count.cast<IntegerType>(kInteger);
	auto cond = condition.cast<BooleanType>(kBoolean);
	if (cond->value) {
		programCounter += jmp->value;
		programCounter--;
	}
}

void ls::VirtualMachine::jf()
{
	auto jmp_count = popValue();
	auto condition = popValue();
	auto jmp = jmp_count.cast<IntegerType>(kInteger);
	auto cond = condition.cast<BooleanType>(kBoolean);
	if (!cond->value) {
		programCounter += jmp->value;
		programCounter--;
	}
}

void ls::VirtualMachine::store()
{
	auto fvar = popValue();//LHS=RHS
	auto lvar = popValue();
	*fvar.val() = *lvar.val();
}

void ls::VirtualMachine::and()
{
	auto LHS = popValue();
	auto RHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->binaryGo(Opers::kAnd, LHS.get()));
	pushValue(tmp);
}

void ls::VirtualMachine::or()
{
	auto LHS = popValue();
	auto RHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->binaryGo(Opers::kOr, RHS.get()));
	pushValue(tmp);
}

void ls::VirtualMachine::not()
{
	auto LHS = popValue();
	TempVariable tmp;
	tmp.val()->set(LHS.get()->unaryGo(Opers::kNot));
	pushValue(tmp);
}

void ls::VirtualMachine::pop()
{
	popValue();
}

void ls::VirtualMachine::push()
{
	TempVariable tmp;
	tmp.setval(&programFrame->operands[programFrame->code[programCounter].pos]);
	pushValue(tmp);
}

void ls::VirtualMachine::inst_ret()
{
	if (callstack.size() == 0)
	{
		throw std::runtime_error("Call Stack Error");
	}
	stack.back().copyTemp();
	auto b = callstack.back();
	programFrame = b.frame;
	programFrameEnd = b.frame->code.size();
	programCounter = b.address;
	callstack.pop_back();
	leaveScope();
	programCounter--;
}

void ls::VirtualMachine::next()
{
	auto LHS = popValue();
	bool vailed = LHS.get()->next();
	if (!vailed) {
		throw std::runtime_error("Type not able to iterate");
	}
	pushValue(LHS);
}

void ls::VirtualMachine::getvalue()
{
	auto LHS = popValue();
	auto t = LHS.cast<TableIterType>(kTableIter);
	TempVariable tmp;
	tmp.val()->set(t->value());
	pushValue(LHS);
	pushValue(tmp);
}

void ls::VirtualMachine::isEnd()
{
	auto LHS = popValue();
	auto t = LHS.cast<TableIterType>(kTableIter);
	TempVariable tmp;
	tmp.val()->set(createVar(t->isEnd()));
	pushValue(LHS);
	pushValue(tmp);
}

void ls::VirtualMachine::getiter()
{
	auto LHS = popValue();
	auto table = LHS.cast<TableType>(kTable);
	TempVariable tmp;
	tmp.val()->set(table->getiter());
	pushValue(tmp);
}

void ls::VirtualMachine::call()
{
	auto f = popValue();
	switch (f.get()->Type()) {
	case kFunctionProxy:
	{
		auto fn = static_cast<FunctionProxyType *>(f.get());
		auto arg_numT = popValue();
		auto arg_num = arg_numT.cast<IntegerType>(kInteger);
		std::vector<BoxedVar> C_stack;
		for (size_t i = 0,endnum = arg_num->value; i < endnum; i++)
		{
			auto tmp = popValue();
			C_stack.push_back(*tmp.val());
		}
		auto ret = TempVariable();
		ret.val()->set(fn->call(C_stack));
		pushValue(ret);
	}
	break;
	case kFunction:
	{
		auto fn = static_cast<FunctionType *>(f.get());
		callstack.push_back({ programFrame,programCounter+1 });
		programFrame = fn->code.get();
		programFrameEnd = programFrame->code.size();
		programCounter = -1;
	}
	break;
	default:
		unexpected();//
		break;
	}
}

void ls::VirtualMachine::ref()
{
	auto LHS = popValue();
	auto RHS = popValue();
	TempVariable tmp;
	tmp.setval(LHS.get()->index(RHS.get()));
	pushValue(tmp);
}
