#include <iostream>
#include <fstream>
#include <list>
#include "Lexer.h"
#include "Parser.h"
#include "VirtualMachine.h"

class Option {
public:
	Option(int n, const char ** v) :argnum(n), argvalue(v),cur(1)
	{}

	const char * find(const char * optname) {
		for (size_t i = cur; i < argnum; i++)
		{
			if (strcmp(argvalue[i], optname) == 0)
			{
				if (i + 1 < argnum)
				{
					return argvalue[i + 1];
				}
				else
				{
					std::cout << "invailed argument";
					exit(1);
				}
			}
		}
		return nullptr;
	}
	const char * getFile() {
		return argvalue[cur++];
	}
private:
	int argnum;
	const char ** argvalue;

	int cur;
};

static void usage()
{
	std::cout <<
		"usage : LightScript inputfile [-m][-d]\n"
		"	-m, --mode : intepreter, lua_script\n"
		"		(default is native)\n"
		"	-d --debug : true, false\n"
		"		(default is false)\n"
		"	-o --output : path of the output file\n"
		"				  only needed if mode is lua_script.\n"
		"		(default is \"./output.lua\")\n";
}

//std::shared_ptr<CompoundType> call(std::vector<BoxedVar> & stack)
std::shared_ptr<ls::CompoundType> my_print(std::vector<ls::BoxedVar> & stack) {
	std::cout << stack[0].to_string() << std::endl;
	return ls::createVar();
}

//=================================
//option
//=================================
bool debug = false;
std::string mode = "native";
std::string output_path = "./output.lua";
//=================================
void setOption(Option & opt);

int main(int argc, const char * argv[]) {
	Option opt(argc, argv);
	if (argc == 1) {
		usage();
		return 0;
	}
	std::string filename(opt.getFile());
	
	setOption(opt);

	std::list<ls::Token> tokens;
	ls::GenerationContext gct;
	ls::VirtualMachine vm;
	{
		std::ifstream file(filename);
		
		ls::IstreamTokenizer tokenizer(file);
		ls::Token t;
		do {
			t = tokenizer.get();
			tokens.push_back(t);
		} while (t.type != ls::ttEof);
		file.close();
	}
	if (debug)
	{
		std::cout << 
			"========================\n"
			"tokens\n"
			"========================\n";
		for (auto it : tokens)
		{
			std::cout << it;
		}
		std::cout << std::endl;
	}
	{
		ls::Parser ps;
		ps.set(tokens.begin());
		std::unique_ptr<ls::AST_SentenceList> tree;
		try
		{
			tree = ps.parseSentenceList();
		}
		catch (const ls::ExpectedException & e)
		{
			std::cout <<"Grammar error : "<< e.what() << "\n"
				" line : " << e.line << "\n";
			return 1;
		}
		
		if (debug)
		{
			std::cout << "========================\n"
				<<"tree\n"
				<<"========================\n"
				<<tree << std::endl;
		}
		if (mode == "native") {
			tree->generateCode(gct);
			if (debug)
			{
				std::cout <<"========================\n"
					"instruction code\n"
					"========================\n"
					<< gct
					<< "========================\n";
			}
		}
		else if(mode == "lua_script") {
			std::ofstream output(output_path);
			if (!output.is_open()) {
				std::cout << "open failed!\n";
				return 1;
			}
			ls::LuaScriptGenerateContext lsgc;
			lsgc.out(output);
			tree->luaCodeGen(lsgc);
			output.close();
			return 0;
		}
		else {
			std::cout << "invailed mode input";
			return 1;
		}
	}
	{
		vm.setVar("print") = ls::createBoxedFunction(my_print);
		vm.set(gct.getFrame());
		try
		{
			while (vm.runOneCode());
		}
		catch (const std::exception& e)
		{
			std::cout << e.what();
			return 1;
		}
	}
	return 0;
}

void setOption(Option & opt) {
	const char * d1 = opt.find("-d");
	if (d1 != nullptr)
	{
		debug |= (strcmp(d1, "true") == 0);
	}
	const char * d2 = opt.find("--debug");
	if (d2 != nullptr)
	{
		debug |= (strcmp(d2, "true") == 0);
	}
	const char * t1 = opt.find("-m");
	const char * t2 = opt.find("--mode");
	if (t1 != nullptr)
	{
		mode = t1;
	}
	if (t2 != nullptr)
	{
		mode = t2;
	}
	t1 = opt.find("-o");
	t2 = opt.find("--ouput");
	if (t1 != nullptr)
	{
		output_path = t1;
	}
	if (t2 != nullptr)
	{
		output_path = t2;
	}
}